# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base

class Dict(dict):
    pass

def build_models_attributes():
    metadata = sqlalchemy.MetaData()
    t1_id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id')
    models_attributes = Dict({
        't1': {
            '__table__': sqlalchemy.Table('t1', metadata, t1_id,
                sqlalchemy.Column(sqlalchemy.Boolean, name='c1', autoincrement=False),
                sqlalchemy.Column(sqlalchemy.Integer, name='c2', autoincrement=False),
                sqlalchemy.Column(sqlalchemy.Float, name='c3'),
                sqlalchemy.Column(sqlalchemy.String(255), name='c4'),
                sqlalchemy.Column(sqlalchemy.Text(), name='c5'),
                sqlalchemy.Column(sqlalchemy.Boolean, name='c6', nullable=True, autoincrement=False),
                sqlalchemy.Column(sqlalchemy.Integer, name='c7', primary_key=False, autoincrement=False),
                sqlalchemy.Column(sqlalchemy.Float, name='c8', primary_key=False),
                sqlalchemy.Column(sqlalchemy.String(255), name='c9', default='test'),
                sqlalchemy.Column(sqlalchemy.Text(), name='c10'),
                sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('t1.id',
                                                             onupdate='CASCADE',
                                                             ondelete='CASCADE'),
                                                                         name='t1_id',
                                                                         autoincrement=False),
                sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('t2.id',
                                                             onupdate='CASCADE',
                                                             ondelete='CASCADE'),
                                                                         name='t2_id',
                                                                         autoincrement=False,
                                                                         primary_key=False)),
            't1': sqlalchemy.orm.relationship('t1', cascade='all', remote_side=[t1_id]),
            't2': sqlalchemy.orm.relationship('t2', cascade='all',
                backref=sqlalchemy.orm.backref('t1', cascade='all,delete-orphan')),
            't3': sqlalchemy.orm.relationship('t3', cascade='all',
                backref=sqlalchemy.orm.backref('t1', cascade='all'),
                                                    secondary='t1_t3_association')
        },
        't2': {
            '__table__': sqlalchemy.Table('t2', metadata,
                sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id'),
                sqlalchemy.Column(sqlalchemy.String, name='c1'))
        },
        't3': {
            '__table__': sqlalchemy.Table('t3', metadata,
                sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id'))
        },
        't4': {
            '__table__': sqlalchemy.Table('t4', metadata,
                sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id'),
                sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('t1.id',
                                                                 onupdate='CASCADE',
                                                                 ondelete='CASCADE'),
                                                                             name='t1_id',
                                                                         autoincrement=False)),
            't1': sqlalchemy.orm.relationship('t1', cascade='all',
                        backref=sqlalchemy.orm.backref('t4', cascade='all,delete-orphan'))
        },
        't5': {
            '__table__': sqlalchemy.Table('t5', metadata,
                sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id')),
            't1': sqlalchemy.orm.relationship('t1', cascade='all',
                    backref=sqlalchemy.orm.backref('t5', cascade='all'),
                                                         secondary='t5_t1_association')
        },
        't6': {
            '__table__': sqlalchemy.Table('t6', metadata,
                sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id')),
            't1': sqlalchemy.orm.relationship('t1', secondary='t6_t1_association',
                                                            cascade='all',
                                                             backref=sqlalchemy.orm.backref('t6', cascade='all')),
            't2': sqlalchemy.orm.relationship('t2', secondary='t6_t2_association',
                                                             cascade='all',
                                                             backref=sqlalchemy.orm.backref('t6', cascade='all'))
        },
        't7': {
            '__table__': sqlalchemy.Table('t7', metadata,
                sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, name='id'))
        }
    })

    sqlalchemy.Table('t6_t2_association', metadata,
    sqlalchemy.Column('t6_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t6.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False),
    sqlalchemy.Column('t2_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t2.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False))

    sqlalchemy.Table('t6_t1_association', metadata,
    sqlalchemy.Column('t6_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t6.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False),
    sqlalchemy.Column('t1_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t1.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False))

    sqlalchemy.Table('t5_t1_association', metadata,
    sqlalchemy.Column('t5_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t5.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False),
    sqlalchemy.Column('t1_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t1.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False))

    sqlalchemy.Table('t1_t3_association', metadata,
    sqlalchemy.Column('t1_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t1.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False),
    sqlalchemy.Column('t3_id', sqlalchemy.Integer, sqlalchemy.ForeignKey('t3.id', onupdate='CASCADE', ondelete='CASCADE'), primary_key=True, autoincrement=False))

    models_attributes.metadata = metadata
    return models_attributes

def build_models():
    models_attributes = build_models_attributes()

    metadata = models_attributes.metadata

    base = declarative_base(metadata=metadata)
    
    models = models_attributes.copy()
    for model_name, table in models.iteritems():
        model = type(model_name, (base,), table)
        models[model_name] = model
    
    for model in models.values():
        model.__mapper__.relationships

    models = Dict(models)
    models.metadata = metadata
    return models

models_expected = build_models()

attributes_expected = build_models_attributes()



# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import pytest
from alquimia.models import AlquimiaModels
from sqlalchemy.testing.engines import mock_engine
from sqlalchemy import MetaData, create_engine

@pytest.fixture(scope='session')
def user_models():
    return {
    't1': {
        'c1': {'type': 'boolean', 'autoincrement': False},
        'c2': {'type': 'integer', 'autoincrement': False},
        'c3': 'float',
        'c4': 'string',
        'c5': 'text',
        'c6': {'type': 'boolean', 'nullable': True, 'autoincrement': False},
        'c7': {'type': 'integer', 'primary_key': False, 'autoincrement': False},
        'c8': 'float',
        'c9': {'type': 'string', 'default': 'test'},
        'c10': {'type': 'text'},
        'relationships': ['t1', {'t2': {'primary_key': False}}, {'t3': 'many-to-many'}]
    },
    't2': {'c1': 'string'},
    't3': {},
    't4': {'relationships': ['t1']},
    't5': {'relationships': {'t1': {'many-to-many': True}}},
    't6': {'relationships': [{'t1': 'many-to-many'}, {'t2': 'many-to-many'}]},
    't7': {}
    }

@pytest.fixture(scope='session')
def db_uri():
    return os.environ.get('ALQUIMIA_TEST_DB') or 'mysql://root:root@localhost:3600/alquimia_test'

def models_finalizer(models_):
    for model in models_.values():
        s = model.session()
        s.query(model).delete()
        s.commit()
        s.close()

@pytest.fixture(scope='session')
def models_create(request, user_models, db_uri):
    metadata = MetaData(db_uri)
    metadata.reflect()
    metadata.drop_all()
    
    models_ = AlquimiaModels(db_uri, user_models, create=True)
    return models_

@pytest.fixture
def models(request, user_models, models_create, db_uri):
    models_ = AlquimiaModels(db_uri, user_models)
    def fin():
        models_finalizer(models_)
    request.addfinalizer(fin)
    return models_

@pytest.fixture
def models_reflect(request, models_create, db_uri):
    models_ = AlquimiaModels(db_uri)
    def fin():
        models_finalizer(models_)
    request.addfinalizer(fin)
    return models_

@pytest.fixture
def t1_t2_query():
    return {
        'c4': 'test1',
        't2': {
            'c1': 'test12'
        }
    }

@pytest.fixture
def t1_simple_obj():
    return {'c1': True, 'c2': 1, 'c4': 'test1'}

@pytest.fixture
def t1_t2_obj(t1_simple_obj):
    obj = t1_simple_obj.copy()
    obj.update({'t2': {'c1': 'test12'}, 't1': {'c4': 'test11'}})
    return obj

@pytest.fixture
def t1_t2_update():
    return {'c4': 'test_updated', 't2': {'c1': 'test_updated'}}



# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.orm.attributes import InstrumentedAttribute
import weakref


class AlquimiaModelMeta(DeclarativeMeta):
    def __init__(cls, classname, bases, dict_):
        DeclarativeMeta.__init__(cls, classname, bases, dict_)
        attrs = {k:v for k,v in cls.__dict__.items() \
                                       if isinstance(v, InstrumentedAttribute)}
        cls.__attrs__ = cls.__attributes__ = attrs
        cls._current_pos = 0

    @property
    def keys(cls):
        return cls.__attrs__.keys

    @property
    def has_key(cls):
        return cls.__attrs__.has_key
    
    @property
    def values(cls):
        return cls.__attrs__.values

    @property
    def items(cls):
        return cls.__attrs__.items
    
    @property
    def iteritems(cls):
        return cls.__attrs__.iteritems
    
    def __getitem__(cls, attr_name):
        try:
            return cls.__attributes__[attr_name]
        except KeyError, e:
            raise KeyError(e.message)
        
    def __iter__(cls):
        return cls

    def next(cls):
        self._current_pos += 1
        if self._current_pos >= len(self.keys()):
            self._current_pos = 0
            raise StopIteration
        else:
            return self.keys()[self._current_pos - 1]

    def _build_objs(cls, obj):
        if not isinstance(obj, list):
            obj = [obj]
        objs = []
        for each in obj:
            if not isinstance(each, cls):
                each = cls(**each)
            objs.append(each)
        return objs

    def insert(cls, objs):
        session = cls.session()
        objs = cls._build_objs(objs)
        session.add_all(objs)
        session.commit()
        return objs

    def _update_rec(cls, obj_dict, obj):
        for prop_name, prop in obj_dict.iteritems():
            if isinstance(prop, dict):
                cls._update_rec(prop, obj[prop_name])
            else:
                obj[prop_name] = prop

    def update(cls, objs):
        session = cls.session()
        if not isinstance(objs, list):
            objs = [objs]
        for obj in objs:
            new_obj = session.query(cls).filter(cls.id == obj.pop('id')).one()
            cls._update_rec(obj, new_obj)
        session.commit()
        return new_obj

    def delete(cls, ids):
        session = cls.session()
        if not isinstance(ids, list):
            ids = [ids]
        for id_ in ids:
            session.query(cls).filter(cls.id == id_).delete()
        session.commit()

    def _build_query_filter_rec(cls, query_dict, obj, filters):
        for prop_name, prop in query_dict.iteritems():
            if isinstance(prop, dict):
                cls._build_query_filter_rec(prop, obj[prop_name], filters)
            else:
                if hasattr(obj, 'model'):
                    obj = obj.model
                filters.append(obj[prop_name] == prop)
        return filters

    def query(cls, query_dict):
        filters = cls._build_query_filter_rec(query_dict, cls, [])
        session = cls.session()
        return session.query(cls).filter(*filters).all()


class AlquimiaModel(object):
    def _check_attr(self, attr_name):
        if not type(self).has_key(attr_name):
            raise TypeError("'%s' is not a valid model attribute!" % attr_name)

    def __init__(self, **kwargs):
        for prop_name, prop in kwargs.iteritems():
            if isinstance(prop, dict):
                self[prop_name] = type(self)[prop_name].model(**prop)
            else:
                self[prop_name] = prop
        self._current_pos = 0

    def __setitem__(self, item, value):
        self._check_attr(item)
        setattr(self, item, value)

    def __getitem__(self, item):
        self._check_attr(item)
        return getattr(self, item)

    def __repr__(self):
        return str(dict(self))

    def __iter__(self):
        return self

    def next(self):
        self._current_pos += 1
        if self._current_pos >= len(self.keys()):
            self._current_pos = 0
            raise StopIteration
        else:
            return self.keys()[self._current_pos - 1]

    def has_key(self, key):
        awld._check_attr(key)
        return hasattr(self, key)

    def keys(self):
        return type(self).keys()


# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
from setuptools import setup

if sys.version_info < (2, 7):
    raise Exception("Alquimia requires Python 2.7 or higher.")
elif sys.version_info >= (3, 0):
    raise Exception("Alquimia do not support Python 3")

setup(
    name='Alquimia',
    version='0.0.5',
    author='Diogo Dutra',
    author_email='dutradda@gmail.com',
    description='An API to work with JSON schemas in SQLAlchemy',
    license='LGPLv3',
    keywords='json sqlachemy sql orm database',
    url='http://packages.python.org/alquimia',
    packages=['alquimia', 'tests'],
    package_data={'alquimia': ['schemas/models.json']},
    requires=['sqlalchemy', 'jsonschema'],
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'
    ]
)



# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os.path
import sqlalchemy
import json

ROOT_DIR = os.path.dirname(os.path.realpath(__file__))

SCHEMAS_DIR = os.path.join(ROOT_DIR, 'schemas')

MODELS_SCHEMA_FILE = os.path.join(SCHEMAS_DIR, 'models.json')

SCHEMA = json.loads(open(MODELS_SCHEMA_FILE).read())

DATA_TYPES = {
    'integer': sqlalchemy.Integer,
    'string': sqlalchemy.String(255),
    'float': sqlalchemy.Float,
    'text': sqlalchemy.Text(),
    'boolean': sqlalchemy.Boolean
}

from alquimia.models import AlquimiaModels



# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import jsonschema
import json
from abc import ABCMeta
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship, sessionmaker
from sqlalchemy import Column, ForeignKey, MetaData, Table
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.orm import ColumnProperty
from alquimia.model import AlquimiaModel, AlquimiaModelMeta
from alquimia import SCHEMA, DATA_TYPES


class ModelsAtrrsReflect(dict):
    __metaclass__ = ABCMeta

    def __init__(self, metadata, *args):
        self._metadata = metadata
        self._rels = {}
        self._build(*args)

    @property
    def metadata(self):
        return self._metadata

    def _build_rel_args(self, rel_name, cascade='all'):
        args = {}
        args['args'] = [rel_name]
        args['cascade'] = cascade
        return args

    def _build_rel_instance(self, args, rel_name, table_name):
        self[table_name][rel_name] = relationship(*args.pop('args'), **args)

    def _raises_one_to_one_error(self, model_name, rel_name):
        raise TypeError('%s.%s is a one-to-one relationship but ' \
                                  'was mapped as many-to-many!' % \
                                            (model_name, rel_name))

    def _build_many_to_many_rel(self, rel_name, table_name, mtm_table):
        if rel_name == table_name:
            self._raises_one_to_one_error(table_name, rel_name)
        self[table_name]['mtm'].append(rel_name)
        self[rel_name]['mtm'].append(table_name)
        self[table_name]['relationships'].append(rel_name)
        self[rel_name]['relationships'].append(table_name)
        args = self._build_rel_args(rel_name)
        args['secondary'] = mtm_table
        args_br = self._build_rel_args(table_name)
        args_br['secondary'] = mtm_table
        self._build_rel_instance(args, rel_name, table_name)
        self._build_rel_instance(args_br, table_name, rel_name) # backref

    def _build_many_to_one_rel(self, rel_name, table_name):
        self[table_name]['mto'].append(rel_name)
        self[rel_name]['otm'].append(table_name)
        self[table_name]['relationships'].append(rel_name)
        self[rel_name]['relationships'].append(table_name)
        args = self._build_rel_args(rel_name)
        args_br = self._build_rel_args(table_name, 'all,delete-orphan')
        self._build_rel_instance(args, rel_name, table_name)
        self._build_rel_instance(args_br, table_name, rel_name)
    
    def _build_one_to_one_rel(self, rel_name, table_name, id_column):
        self[table_name]['oto'].append(rel_name)
        self[table_name]['relationships'].append(rel_name)
        args = self._build_rel_args(rel_name)
        args['uselist'] = False
        args['remote_side'] = [id_column]
        self._build_rel_instance(args, rel_name, table_name)

    def _build_relationships(self, table):
        for fk in table.foreign_keys:
            rel_name = fk.column.table.name
            if rel_name == table.name:
                self._build_one_to_one_rel(rel_name, table.name,
                                                                 table.c['id'])
            else:
                self._build_many_to_one_rel(rel_name, table.name)
        mtm_tables = list(self._mtm_tables.values())
        for mtm_table in mtm_tables:
            mtm_rels = list(mtm_table.columns.keys())
            table_rel = table.name+'_id'
            if table_rel in mtm_rels:
                mtm_rels.remove(table_rel)
                rel_name = mtm_rels[0][:-3]
                self._build_many_to_many_rel(rel_name, table.name,
                                                                mtm_table.name)
                self._mtm_tables.pop(mtm_table.name)

    def _keep_mtm_tables(self):
        self._mtm_tables = {}
        for table in self._metadata.tables.values():
            fks = table.foreign_keys
            if len(fks) == len(table.c) == 2:
                is_pks = True
                for fk in fks:
                    if not fk.column.primary_key:
                        is_pks = False
                        break
                if not is_pks:
                    continue
                self._mtm_tables[table.name] = table

    def _init_attrs(self, table_name):
        self[table_name] = {
            'mtm': [],
            'mto': [],
            'oto': [],
            'otm': [],
            'relationships': [],
            'columns': [],
            'session': None
        }

    def _build(self, *args):
        self._metadata.reflect()
        self._keep_mtm_tables()
        attrs = {}
        tables = [table for table in self._metadata.tables.values() \
                                         if table.name not in self._mtm_tables]
        for table in tables:
            self._init_attrs(str(table.name))

        for table in tables:
            if table.name not in self._mtm_tables:
                self._build_relationships(table)
                self[table.name]['__table__'] = table


class ModelsAttributes(ModelsAtrrsReflect):
    def __init__(self, dict_, metadata, data_types=DATA_TYPES):
        self._data_types = data_types
        ModelsAtrrsReflect.__init__(self, metadata, *[dict_])

    def _build_columns(self, model_name, model):
        new_model = {}
        for k, v in model.iteritems():
            new_model[k] = type(v)(v)
        model = new_model
        model.pop('relationships', None)
        model['id'] = {'type': 'integer', 'primary_key': True}
        for col_name, column in model.iteritems():
            if not isinstance(column, dict):
                type_ = column
                column = {'args': [col_name, type_]}
            else:
                type_ = column.pop('type', None)
                if not type_:
                    raise TypeError('Column %s.%s type must be defined!' \
                                                      % (model_name, col_name))
                column['args'] = [col_name, type_]
            self._build_column_instance(column, col_name, model_name)

    def _build_relationships_dict(self, rels):
        new_rels = {}
        if not isinstance(rels, dict):
            for rel in rels:
                if not isinstance(rel, dict):
                    new_rels[rel] = {}
                elif rel.values()[0] == 'many-to-many':
                    new_rels[rel.keys()[0]] = {'many-to-many': True}
                elif rel.values()[0] == 'primary_key':
                    new_rels[rel.keys()[0]] = {'primary_key': True}
                else:
                    new_rels[rel.keys()[0]] = rel.values()[0]
        else:
            for k, v in rels.iteritems():
                new_rels[k] = type(v)(v)
        return new_rels

    def _build_relationship_column(self, rel_name, model_name, primary_key):
        foreign_key = {
            'args': [rel_name+'.id'],
            'onupdate': 'CASCADE',
            'ondelete': 'CASCADE'
        }
        rel_col_name = rel_name+'_id'
        column = {'args': [rel_col_name, 'integer', foreign_key],
                  'autoincrement': False}
        if primary_key:
            column['primary_key'] = primary_key
        self._build_column_instance(column, rel_col_name, model_name)

    def _build_relationships(self, model_name, rels_model):
        rels_dict = self._build_relationships_dict(rels_model)
        rels = {}
        for rel_name, rel in rels_dict.iteritems():
            if self.has_key(rel_name+'_id'):
                raise TypeError(rel+'_id is a alquimia primitive, '\
                                                               "don't use it!")
            mtm_table_name = rel.pop('many-to-many', None)
            if mtm_table_name is not None:
                mtm_table_name = '%s_%s_association' % (model_name, rel_name)
                self._build_many_to_many_table(model_name, rel_name,
                                                                mtm_table_name)
                self._build_many_to_many_rel(rel_name,
                                                    model_name, mtm_table_name)
            else:
                primary_key = rel.pop('primary_key', False)
                self._build_relationship_column(rel_name, model_name,
                                                                   primary_key)
                if rel_name == model_name:
                    self._build_one_to_one_rel(rel_name, model_name,
                                                        self[model_name]['id'])
                else:
                    self._build_many_to_one_rel(rel_name, model_name)
        return rels

    def _build_column_instance(self, column, col_name, model_name):
        if len(column['args']) == 3:
            fk = column['args'][2]
            column['args'][2] = ForeignKey(*fk.pop('args'), **fk)
        column['args'][1] = self._data_types[column['args'][1]]
        self[model_name][col_name] = Column(*column.pop('args'), **column)

    def _build_many_to_many_table(self, rel1_name, rel2_name, table_name):
        col1 = Column(rel2_name+'_id', self._data_types['integer'],
           ForeignKey(rel2_name+'.id', onupdate='CASCADE', ondelete='CASCADE'),
           primary_key=True, autoincrement=False)
        col2 = Column(rel1_name+'_id', self._data_types['integer'],
           ForeignKey(rel1_name+'.id', onupdate='CASCADE', ondelete='CASCADE'),
           primary_key=True, autoincrement=False)
        Table(table_name, self._metadata, col1, col2)

    def _build(self, dict_):
        for model_name in dict_.keys():
            self._init_attrs(model_name)

        for model_name, model in dict_.iteritems():
            self._build_columns(model_name, model)
            rels = model.get('relationships', {})
            self._build_relationships(model_name, rels)
            self[model_name]['__tablename__'] = model_name


class AlquimiaModels(dict):
    def __init__(self, db_url, dict_=None, data_types=DATA_TYPES,
                                                    engine=None, create=False):
        if not engine:
            engine = create_engine(db_url)
        base_model = declarative_base(engine, metaclass=AlquimiaModelMeta,
                         cls=AlquimiaModel, constructor=AlquimiaModel.__init__)
        self._session = sessionmaker(engine)
        self._metadata = base_model.metadata
        if dict_ is not None:
            attrs = ModelsAttributes(dict_, self._metadata, data_types)
        else:
            attrs = ModelsAtrrsReflect(self._metadata)
        self._build(base_model, attrs)
        if create:
            self._metadata.create_all()

    @property
    def metadata(self):
        return self._metadata

    @property
    def session(self):
        return self._session
    
    def _set_relationship(self, rel):
        pass

    def _build(self, base_model, models_attrs):
        models = {}
        for model_name, attrs in models_attrs.iteritems():
            attrs.update({'session': self._session})
            model = type(model_name, (base_model,), attrs)
            models[model_name] = model

        for model in models.values():
            model.__mapper__.relationships
            for attr_name, attr in model.iteritems():
                if isinstance(attr.prop, RelationshipProperty):
                    setattr(attr, 'model', models[attr_name])
                elif isinstance(attr.prop, ColumnProperty):
                    model.columns.append(attr_name)

        self.update(models)



# Copyright 2015 Diogo Dutra

# This file is part of Alquimia.

# Alquimia is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pytest
import sqlalchemy
import json
from tests.models_expected import models_expected

class TestAlquimiaModels(object):
    def _check_foreign_key(self, fk1, fk2):
        assert len(fk1) == len(fk2)
        assert not len(fk1) > 1
        if len(fk1) == 1:
            fk1 = list(fk1)[0]
            fk2 = list(fk2)[0]
            assert fk1.onupdate == fk2.onupdate
            assert fk1.ondelete == fk2.ondelete
            self._check_column(fk1.column, fk2.column, False)
            self._check_column(fk1.parent, fk2.parent, False)

    def _check_column(self, column1, column2, test_fk=True, reflect=False):
        assert column1.name == column2.name
        c1_type = column1.type.python_type
        c2_type = column2.type.python_type
        if not ((c1_type == int and c2_type == bool) or \
                                         (c2_type == int and c1_type == bool)):
            assert column1.type.python_type == column2.type.python_type
        assert column1.primary_key == column2.primary_key
        assert column1.nullable == column2.nullable
        assert bool(column1.autoincrement) == bool(column2.autoincrement)
        assert column1.unique == column2.unique
        if isinstance(column1.default, sqlalchemy.schema.ColumnDefault):
            assert column1.default.arg == column2.default.arg
        elif column1.default is not None:
            assert column1.default == column2.default
        if test_fk:
            self._check_foreign_key(column1.foreign_keys, column2.foreign_keys)

    def _check_tables(self, tables, tables_exp):
        for i in range(len(tables)):
            for col_name, col in tables[i].columns.items():
                col_exp = tables_exp[i].columns[col_name]
                self._check_column(col, col_exp)

    def _check_relationship(self, rel1, rel2):
        rel1.table.name == rel2.table.name
        assert rel1.lazy == rel2.lazy
        assert rel1.innerjoin == rel2.innerjoin
        assert rel1.order_by == rel2.order_by
        rs = (rel1.remote_side and rel2.remote_side)
        assert rs is None or rs
        if rs and len(rel1.remote_side) == len(rel2.remote_side) == 1:
            self._check_column(list(rel1.remote_side)[0], list(rel2.remote_side)[0])
        assert rel1.cascade == rel2.cascade
        if rel1.secondary is not None and rel2.secondary is not None:
            self._check_tables([rel1.secondary], [rel2.secondary])

    def _check_models(self, models):
        tables = models.metadata.sorted_tables
        tables_exp = models_expected.metadata.sorted_tables
        assert len(tables) == len(tables_exp)
        self._check_tables(tables, tables_exp)

        for model_name, model in models.iteritems():
            rels = model.__mapper__.relationships
            rels_exp = models_expected[model_name].__mapper__.relationships
            assert len(rels) == len(rels_exp) == len(model.relationships)
            for rel_name in model.relationships:
                rel_exp = getattr(models_expected[model_name], rel_name).prop
                self._check_relationship(model[rel_name].prop, rel_exp)

    def test_create(self, models_create):
        self._check_models(models_create)

    def test_models(self, models):
        self._check_models(models)
    
    def test_models_reflect(self, models_reflect):
        self._check_models(models_reflect)


class TestAlquimiaModel(object):
    def test_init_model(self, models, t1_simple_obj):
        obj = models['t1'](**t1_simple_obj)
        assert obj['c1'] == t1_simple_obj['c1']
        assert obj['c2'] == t1_simple_obj['c2']
        assert obj['c4'] == t1_simple_obj['c4']

    def test_init_rec(self, models, t1_t2_obj):
        obj = models['t1'](**t1_t2_obj)
        assert type(obj['t2']) == models['t2']
        assert obj['t2']['c1'] == t1_t2_obj['t2']['c1']

    def test_model_repr(self, models, t1_t2_obj):
        obj = models['t1'].insert(t1_t2_obj)
        t1 = dict(obj[0])
        t1['t2'] = dict(t1['t2'])
        assert repr(t1) == repr(obj[0])


class TestAlquimiaModelMeta(object):
    def test_modelmeta_insert(self, models, t1_t2_obj):
        model = models['t1']
        model.insert(t1_t2_obj)
        s = model.session()
        nobj = s.query(model).all()[1]
        assert nobj['c1'] == t1_t2_obj['c1']
        assert nobj['c2'] == t1_t2_obj['c2']
        assert nobj['c4'] == t1_t2_obj['c4']
        assert nobj['t2']['c1'] == t1_t2_obj['t2']['c1']
            
    def test_modelmeta_delete(self, models, t1_t2_obj):
        models['t1'].insert(t1_t2_obj)
        model = models['t1']
        s = model.session()
        assert len(s.query(model).all()) == 2
        obj = s.query(model).filter(model.c4 == 'test11').one()
        model.delete(obj.id)
        s.close()
        s = model.session()
        assert len(s.query(model).all()) == 0

    def test_modelmeta_update(self, models, t1_t2_obj, t1_t2_update):
        model = models['t1']
        model.insert(t1_t2_obj)
        s = model.session()
        id_ = s.query(models['t1'].id).filter(models['t1'].c4 == 'test1').one()[0]
        t1_t2_update.update({'id': id_})
        model.update(t1_t2_update)
        s = model.session()
        obj = s.query(model).filter(model.id == id_).one()
        assert obj['c4'] == t1_t2_update['c4']
        assert obj['t2']['c1'] == t1_t2_update['t2']['c1']

    def test_modelmeta_query(self, models, t1_t2_query, t1_t2_obj):
        models['t1'].insert(t1_t2_obj)
        query = models['t1'].query(t1_t2_query)
        assert len(query) == 1
        query = query[0]
        assert query['c4'] == t1_t2_query['c4']
        assert query['t2']['c1'] == t1_t2_query['t2']['c1']

    def test_modelmeta_query_4(self, models, t1_t2_query, t1_t2_obj):
        models['t1'].insert(t1_t2_obj)
        models['t1'].insert(t1_t2_obj)
        models['t1'].insert([t1_t2_obj, t1_t2_obj])
        query = models['t1'].query(t1_t2_query)
        assert len(query) == 4
        query = query[0]
        assert query['c4'] == t1_t2_query['c4']
        assert query['t2']['c1'] == t1_t2_query['t2']['c1']
